@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2019
+# Copyright (c) 2012-2020
 #
 # Author(s):
 #
@@ -160,13 +160,13 @@ import yaml
 
 import inotify.adapters
 from inotify.constants import IN_MODIFY, IN_CLOSE_WRITE, IN_CREATE, IN_MOVED_TO
-from six.moves.urllib.parse import urlsplit, urlunsplit
-
+from six.moves.urllib.parse import urlsplit, urlunsplit, urlparse
+import socket
 from posttroll.listener import ListenerContainer
 from posttroll.publisher import NoisyPublisher
 from posttroll.message import Message
 from trollmoves.movers import move_it
-from trollmoves.utils import clean_url
+from trollmoves.utils import (clean_url, is_file_local)
 from trollsift import compose
 
 logger = logging.getLogger(__name__)
@@ -268,6 +268,8 @@ class Dispatcher(Thread):
         self.topics = None
         self.listener = None
         self.publisher = None
+        self.host = socket.gethostname()
+
         if publish_port is not None:
             self.publisher = NoisyPublisher(
                 "dispatcher", port=publish_port,
@@ -318,7 +320,17 @@ class Dispatcher(Thread):
                     continue
                 destinations = self.get_destinations(msg)
                 if destinations:
-                    success = dispatch(msg.data['uri'], destinations)
+                    # Check if the url are on another host:
+                    url = urlparse(msg.data['uri'])
+                    if not is_file_local(url):
+                        # This warning may appear even if the file path
+                        # does exist on both servers (a central file server
+                        # reachable from both). But in those cases one
+                        # should probably not use an url scheme but just a
+                        # file path:
+                        raise NotImplementedError(("uri is pointing to a file path on another server! "
+                                                   "Host=<%s> uri netloc=<%s>", self.host, url.netloc))
+                    success = dispatch(url.path, destinations)
                     if self.publisher:
                         self._publish(msg, destinations, success)
 
@@ -361,13 +373,13 @@ class Dispatcher(Thread):
 
     def create_dest_url(self, msg, client, disp_config):
         """Create the destination URL and the connection parameters."""
-        defaults = self.config[client]
+        defaults = self.config[client].copy()
+        if 'filepattern' not in defaults:
+            source_filename = os.path.basename(urlsplit(msg.data['uri']).path)
+            defaults['filepattern'] = source_filename
         info_dict = dict()
         for key in ['host', 'directory', 'filepattern']:
-            try:
-                info_dict[key] = disp_config[key]
-            except KeyError:
-                info_dict[key] = defaults[key]
+            info_dict[key] = disp_config.get(key, defaults[key])
         connection_parameters = disp_config.get(
             'connection_parameters',
             defaults.get('connection_parameters'))
