@@ -38,14 +38,16 @@ LOG_FORMAT = "[%(asctime)s %(levelname)-8s %(name)s] %(message)s"
 
 
 class MoveItBase(object):
+    """Base class for Trollmoves."""
 
     def __init__(self, cmd_args, chain_type):
+        """Initialize the class."""
         self.cmd_args = cmd_args
         self.chain_type = chain_type
         self.running = False
         self.notifier = None
         self.watchman = None
-        self.pub = None
+        self.sync_pub = None
         self._np = None
         self.chains = {}
         setup_logging(cmd_args, chain_type)
@@ -53,26 +55,34 @@ class MoveItBase(object):
         self.setup_watchers(cmd_args)
 
     def reload_cfg_file(self, filename, *args, **kwargs):
+        """Reload configuration file."""
         if self.chain_type == "client":
             from trollmoves.client import reload_config
-            reload_config(filename, self.chains, *args, pub_instance=self.pub,
+            reload_config(filename, self.chains, *args, sync_pub_instance=self.sync_pub,
                           **kwargs)
         else:
             # Also Mirror uses the reload_config from the Server
             from trollmoves.server import reload_config
-            reload_config(filename, self.chains, *args, publisher=self.pub,
-                          **kwargs)
+            reload_config(filename, self.chains, *args, publisher=self.sync_pub,
+                          use_watchdog=self.cmd_args.watchdog,
+                          disable_backlog=self.cmd_args.disable_backlog)
 
     def signal_reload_cfg_file(self, *args):
+        """Handle reload signal."""
         del args
         if self.chain_type == "client":
             from trollmoves.client import reload_config
+            reload_config(self.cmd_args.config_file, self.chains,
+                          sync_pub_instance=self.sync_pub)
         else:
             from trollmoves.server import reload_config
-        reload_config(self.cmd_args.config_file, self.chains,
-                      pub_instance=self.pub)
+            reload_config(self.cmd_args.config_file, self.chains,
+                          publisher=self.sync_pub,
+                          use_watchdog=self.cmd_args.watchdog,
+                          disable_backlog=self.cmd_args.disable_backlog)
 
     def chains_stop(self, *args):
+        """Stop all transfer chains."""
         del args
         if self.chain_type == "client":
             from trollmoves.client import terminate
@@ -87,6 +97,7 @@ class MoveItBase(object):
         terminate(self.chains)
 
     def setup_watchers(self, cmd_args):
+        """Set up watcher for the configuration file."""
         mask = (pyinotify.IN_CLOSE_WRITE |
                 pyinotify.IN_MOVED_TO |
                 pyinotify.IN_CREATE)
@@ -99,18 +110,9 @@ class MoveItBase(object):
         self.notifier = pyinotify.ThreadedNotifier(self.watchman, event_handler)
         self.watchman.add_watch(os.path.dirname(cmd_args.config_file), mask)
 
-    def run(self):
-        signal.signal(signal.SIGTERM, self.chains_stop)
-        signal.signal(signal.SIGHUP, self.signal_reload_cfg_file)
-        self.notifier.start()
-        self.running = True
-        while self.running:
-            time.sleep(1)
-            self.pub.heartbeat(30)
-
 
 def setup_logging(cmd_args, chain_type):
-    """Setup logging"""
+    """Set up logging."""
     global LOGGER
     LOGGER = logging.getLogger('')
     if cmd_args.verbose:
